"""CoAP Air Client."""

# pylint: disable=invalid-name, missing-class-docstring, missing-function-docstring

import binascii
import hashlib
import json
import logging
import os
from abc import ABC, abstractmethod
from collections import OrderedDict

from coapthon import defines
from coapthon.client.helperclient import HelperClient
from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad, unpad


class WrongDigestException(Exception):
    pass


class NotSupportedException(Exception):
    pass


class HTTPAirClientBase(ABC):
    def __init__(self, host, port, debug=False):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel("WARN")
        self.server = host
        self.port = port
        self.debug = debug

    def get_status(self, debug=False):
        if debug:
            self.logger.setLevel("DEBUG")
        status = self._get()
        return status

    def set_values(self, values, debug=False):
        if debug:
            self.logger.setLevel("DEBUG")

        result = True
        for key in values:
            result = result and self._set(key, values[key])

        return result

    @abstractmethod
    def _get(self):
        pass

    @abstractmethod
    def _set(self, key, value):
        pass

    def get_firmware(self):
        status = self._get()
        return status

    def get_filters(self):
        status = self._get()
        return status

    def get_wifi(self):
        raise NotSupportedException

    def set_wifi(self, ssid, pwd):
        raise NotSupportedException


class CoAPAirClient(HTTPAirClientBase):
    SECRET_KEY = "JiangPan"

    def __init__(self, host, port=5683, debug=False):
        super().__init__(host, port, debug)
        self.client = self._create_coap_client(self.server, self.port)
        self._sync()

    def __del__(self):
        # TODO call a close method explicitly instead
        self.client.stop()

    def _create_coap_client(self, host, port):
        return HelperClient(server=(host, port))

    def _sync(self):
        self.syncrequest = binascii.hexlify(os.urandom(4)).decode("utf8").upper()
        self.client_key = self.client.post("/sys/dev/sync", self.syncrequest).payload

    def _decrypt_payload(self, encrypted_payload):
        encoded_counter = encrypted_payload[0:8]
        aes = self._handle_AES(encoded_counter)
        encoded_message = encrypted_payload[8:-64].upper()
        digest = encrypted_payload[-64:]
        calculated_digest = self._create_digest(encoded_counter, encoded_message)
        if digest != calculated_digest:
            raise WrongDigestException
        decoded_message = aes.decrypt(bytes.fromhex(encoded_message))
        unpaded_message = unpad(decoded_message, 16, style="pkcs7")
        return unpaded_message.decode("utf8")

    def _encrypt_payload(self, payload):
        self._update_client_key()
        aes = self._handle_AES(self.client_key)
        paded_message = pad(bytes(payload.encode("utf8")), 16, style="pkcs7")
        encoded_message = binascii.hexlify(aes.encrypt(paded_message)).decode("utf8").upper()
        digest = self._create_digest(self.client_key, encoded_message)
        return self.client_key + encoded_message + digest

    def _create_digest(self, id, encoded_message):
        digest = (
            hashlib.sha256(bytes((id + encoded_message).encode("utf8")))
            .hexdigest()
            .upper()
        )
        return digest

    def _update_client_key(self):
        self.client_key = "{:x}".format(int(self.client_key, 16) + 1).upper()

    def _handle_AES(self, id):
        key_and_iv = hashlib.md5((self.SECRET_KEY + id).encode()).hexdigest().upper()
        half_keylen = len(key_and_iv) // 2
        secret_key = key_and_iv[0:half_keylen]
        iv = key_and_iv[half_keylen:]
        return AES.new(
            bytes(secret_key.encode("utf8")), AES.MODE_CBC, bytes(iv.encode("utf8"))
        )

    def _get(self):
        path = "/sys/dev/status"
        decrypted_payload = None

        try:
            request = self.client.mk_request(defines.Codes.GET, path)
            request.observe = 0
            response = self.client.send_request(request, None, 2)
            encrypted_payload = response.payload
            decrypted_payload = self._decrypt_payload(encrypted_payload)
        except WrongDigestException:
            print("Message from device got corrupted")
        except Exception as e:
            print("Unexpected error:{}".format(e))

        if decrypted_payload is not None:
            return json.loads(decrypted_payload, object_pairs_hook=OrderedDict)[
                "state"
            ]["reported"]
        else:
            return {}

    def _set(self, key, value):
        path = "/sys/dev/control"
        try:
            payload = {
                "state": {
                    "desired": {
                        "CommandType": "app",
                        "DeviceId": "",
                        "EnduserId": "",
                        key: value,
                    }
                }
            }
            encrypted_payload = self._encrypt_payload(json.dumps(payload))
            response = self.client.post(path, encrypted_payload)
            if self.debug:
                print(response)
            return response.payload == '{"status":"success"}'
        except Exception as e:
            print("Unexpected error:{}".format(e))



STATUS_TRANSFORMER = {
    "name" : ("Name: {}", None),
    "type" : ("Type: {}", None),
    "modelid" : ("ModelId: {}", None),
    "swversion" : ("Version: {}", None),
    "StatusType" : ("StatusType: {}", None),
    "ota" : ("Over the air updates: {}", None),
    "Runtime" : ("Runtime: {} hours", lambda runtime: round(((runtime/(1000*60*60))%24), 2)),
    "pwr" : ("Power: {}", lambda pwr: {'1': 'ON', '0': 'OFF'}.get(pwr, pwr)),
    "pm25" : ("PM25: {}", None),
    "rh" : ("Humidity: {}", None),
    "rhset" : ("Target humidity: {}", None),
    "iaql" : ("Allergen index: {}", None),
    "tvoc" : ("Total volatile organic compounds: {}", None),
    "temp" : ("Temperature: {}", None),
    "func" : ("Function: {}", lambda func: {'P': 'Purification', 'PH': 'Purification & Humidification'}.get(func, func)),
    "mode" : ("Mode: {}", lambda mode: {'P': 'auto', 'A': 'allergen', 'S': 'sleep', 'M': 'manual', 'B': 'bacteria', 'N': 'night'}.get(mode, mode)),
    "om" : ("Fan speed: {}", lambda om: {'s': 'silent', 't': 'turbo', 'a': 'auto'}.get(om, om)),
    "aqil" : ("Light brightness: {}", None),
    "aqit" : ("Air quality notification threshold: {}", None),
    "uil" : ("Buttons light: {}", lambda uil: {'1': 'ON', '0': 'OFF'}.get(uil, uil)),
    "ddp" : ("Used index: {}", lambda ddp: {'3': 'Humidity', '1': 'PM2.5', '0': 'IAI'}.get(ddp, ddp)),
    "wl" : ("Water level: {}", None),
    "cl" : ("Child lock: {}", None),
    "dt" : ("Timer: {} hours", lambda dt: None if dt == 0 else dt),
    "dtrs" : ("Timer: {} minutes left", lambda dtrs: None if dtrs == 0 else dtrs),
    "fltt1" : ("HEPA filter type: {}", lambda fltt1: {'A3': 'NanoProtect Filter Series 3 (FY2422)'}.get(fltt1, fltt1)),
    "fltt2" : ("Active carbon filter type: {}", lambda fltt2: {'C7': 'NanoProtect Filter AC (FY2420)'}.get(fltt2, fltt2)),
    "fltsts0" : ("Pre-filter and Wick: clean in {} hours", None),
    "fltsts1" : ("HEPA filter: replace in {} hours", None),
    "fltsts2" : ("Active carbon filter: replace in {} hours", None),
    "wicksts" : ("Wick filter: replace in {} hours", None),
    "err" : ("[ERROR] Message: {}", lambda err: None if err == 0 else {49408: 'no water', 32768: 'water tank open', 49155: 'pre-filter must be cleaned'}.get(err, err)),
}



#!/usr/bin/env python3
import os
import re
import json
import email
import struct
import base64
import hashlib
import binascii
import argparse
import configparser
import urllib.request
import urllib.parse

from pyairctrl.http_client import HTTPAirClient

def parse_attr(str, key):
    p = re.compile('%s=\"(.+?)\"' % key)
    m = p.search(str)
    return m.group(1)


class CloudClient:

    def __init__(self, device_id, debug=False):
        self._device_id = device_id
        self._debug = debug

    def _login(self):
        print('Logging in with {}'.format(self._client_id))
        url = 'https://www.ecdinterface.philips.com/DevicePortalICPRequestHandler/RequestHandler.ashx'
        auth = 'CBAuth Type="TLS", Client="%s", Authentication="%s"' % (self._client_id, self._client_key)
        req_body = {'correlationId': 1}
        req_body['clientHeaders'] = {'clientId': self._client_id,
                                     'deviceType': 'FI-AIR_KPSPROV',
                                     'country': 'US',
                                     'language': 'en',
                                     'components': [{'id': 'FI-AIR-AND', 'version': 64}],
                                     'icpVersion': 'CPPDC_8.2.4_B.2'}
        req_body['body'] = [{'action': {'name': 'PortalLocationRequest', 'version': 1},
                             'requestNr': 1,
                             'errorCode': 0,
                             'content': {'type': 0, 'data': '{\"pageSize\":4}'},
                             'attachResponse': True}]

        req = urllib.request.Request(url=url, data=json.dumps(req_body).encode('ascii'), method='POST')
        req.add_header('Authorization', auth)
        req.add_header('Content-Type', 'application/CB-Message; encoding=JSON')
        with urllib.request.urlopen(req) as response:
            www_auth = response.getheader('WWW-Authenticate')
            if self._debug:
                print(www_auth)
            self._sso_key = parse_attr(www_auth, "SSOKey")
            self._sso_token = parse_attr(www_auth, "SSOToken")
            self._nonce = parse_attr(www_auth, "Nonce")

    def _create_auth(self, url):
        num = 1
        host = urllib.parse.urlparse(url).hostname
        m = hashlib.sha1()
        m.update(b'\x05')
        m.update(self._client_id.encode('ascii'))
        m.update(host.encode('ascii'))
        m.update(base64.b64decode(self._sso_key))
        m.update(struct.pack('<I', num))
        m.update(base64.b64decode(self._nonce))
        key = m.digest()[:16]
        auth0 = base64.b64encode(key).decode('ascii')
        return 'CBAuth Type="SSO", Client="%s", RequestNr="%d", Nonce="%s", SSOToken="%s", Authentication="%s"' % (self._client_id, num, self._nonce, self._sso_token, auth0)

    def _create_account(self):
        print('Creating cloud account ...')
        # hardcoded credentials for account provisioning
        self._client_id = '000000fff0000019'
        self._client_key = 'QR3FiHoEQcSZ9S5XxsZwXQ=='
        self._login()

        url = 'https://kps.dc1.philips.com/KpsRequestHandler/index.ashx'
        auth = self._create_auth(url)

        rand_id = binascii.hexlify(os.urandom(15)).decode('ascii')
        body_data = {'Device': {'Type': 'fira', 'Version': '6.0.1_24', 'Id': rand_id},
                     'Application': {'Type': 'FI-AIR-AND', 'Version': 'com.xakcop.air', 'Id': '64'}}
        req_body = {'correlationId': 1}
        req_body['clientHeaders'] = {'clientId': self._client_id,
                                     'deviceType': 'FI-AIR_KPSPROV',
                                     'country': 'US',
                                     'language': 'en',
                                     'components': [{'id': 'FI-AIR-AND', 'version': 64}],
                                     'icpVersion': 'CPPDC_8.2.4_B.2'}
        req_body['body'] = [{'action': {'name': 'ProvisionRequest', 'version': 1},
                             'requestNr': 1,
                             'errorCode': 0,
                             'content': {'type': 0, 'data': json.dumps(body_data)},
                             'attachResponse': True}]

        req = urllib.request.Request(url=url, data=json.dumps(req_body).encode('ascii'), method='POST')
        req.add_header('Authorization', auth)
        req.add_header('Content-Type', 'application/CB-Message; encoding=JSON')
        with urllib.request.urlopen(req) as response:
            content_type = response.getheader('Content-Type')
            body = response.read().decode('ascii')
            parts = "Content-Type: %s\n\n%s" % (content_type, body)

        if self._debug:
            print(parts)
        msg = email.message_from_string(parts)
        if not msg.is_multipart():
            raise Exception('Expected multipart response')
        for part in msg.get_payload():
            if part.get_content_type() == 'application/cb-message':
                continue
            p = json.loads(part.get_payload(decode=False))
            client_id, client_key = p['ClientId'], p['Key']
            bytes_key = bytes.fromhex(client_key)
            client_key = base64.b64encode(bytes_key).decode('ascii')

        print('Client id: {}'.format(client_id))
        print('Client key: {}'.format(client_key))
        config = configparser.ConfigParser()
        fpath = os.path.expanduser('~/.pyairctrl')
        config.read(fpath)
        if 'cloud' not in config.sections():
            config['cloud'] = {}
        config['cloud']['client_id'] = client_id
        config['cloud']['client_key'] = client_key
        self._client_id = client_id
        self._client_key = client_key
        with open(fpath, 'w') as f:
            config.write(f)

    def load_credentials(self):
        fpath = os.path.expanduser('~/.pyairctrl')
        if os.path.isfile(fpath):
            config = configparser.ConfigParser()
            config.read(fpath)
            if 'cloud' in config and 'client_id' in config['cloud']:
                self._client_id = config['cloud']['client_id']
                self._client_key = config['cloud']['client_key']
            else:
                self._create_account()
        else:
            self._create_account()

    def _multi_part(self, part1, part2):
        part1_str = json.dumps(part1).replace(' ', '')
        part2_str = json.dumps(part2).replace(' ', '')
        result = "--ICPMimeBoundary\r\n"
        result += "Content-Type: application/CB-Message; encoding=JSON\r\n"
        result += "Content-Length: " + str(len(part1_str)).zfill(10) + "\r\n\r\n"
        result += part1_str + "\r\n\r\n"
        result += "--ICPMimeBoundary\r\n"
        result += "Content-Type: application/octet-stream\r\n"
        result += "Content-Length: " + str(len(part2_str)).zfill(10) + "\r\n\r\n"
        result += part2_str + "\r\n\r\n"
        result += "--ICPMimeBoundary--"
        return result.encode('ascii')

    def set_values(self, values):
        self._login()

        print('Sending event {} to device with id {}'.format(str(values), self._device_id))

        url = 'https://ep.dcs.dc1.philips.com/DCS.EventPublisherService/Services/External/json'
        auth = self._create_auth(url)

        part1 = {}
        part1['correlationId'] = 14
        part1['clientHeaders'] = {'clientId': self._client_id,
                                  'deviceType': 'FI-AIR-AND',
                                  'country': 'US',
                                  'language': 'en',
                                  'components': [{'id': 'FI-AIR-AND', 'version': 64}]}
        part1['body'] = [{'action': {'name': 'PublishEventRequest', 'version': 1},
                          'requestNr': 1,
                          'errorcode': 0,
                          'content': {'type': 1, 'attachment': {}},
                          'attachResponse': True}]

        part2_data = {'product': '1',
                      'port': 'air',
                      'data': values}
        part2 = {}
        part2['event'] = {'eventType': 'DICOMM-REQUEST',
                          'priority': 20,
                          'replyTo': self._client_id,
                          'conversationId': '',
                          'serviceTag': '',
                          'data': json.dumps(part2_data),
                          'action': 'PUTPROPS'}
        part2['targets'] = [self._device_id]
        part2['ttl'] = 5

        req_body = self._multi_part(part1, part2)

        req = urllib.request.Request(url=url, data=req_body, method='POST')
        req.add_header('Authorization', auth)
        req.add_header('Content-Type', 'multipart/mixed; boundary="ICPMimeBoundary"')
        with urllib.request.urlopen(req) as response:
            resp = response.read().decode('ascii')
            if self._debug:
                print(resp)

    def pair(self, ip_addr):
        cc = HTTPAirClient(ip_addr)
        cc.load_key()
        print('Pairing with {} ...'.format(ip_addr))
        pair_secret = 'ad388b4036986421'
        cc.pair(self._client_id, pair_secret)

        self._login()
        print('Sending pair request to cloud service ...')

        url = 'http://ps.dc1.philips.com/PSRequestHandler/index.ashx'
        auth = self._create_auth(url)

        part1 = {}
        part1['correlationId'] = 15
        part1['clientHeaders'] = {'clientId': self._client_id,
                                  'deviceType': 'FI-AIR-AND',
                                  'country': 'US',
                                  'language': 'en',
                                  'components': [{'id': 'FI-AIR-AND', 'version': 64}]}
        part1['body'] = [{'action': {'name': 'AddRelationshipRequest', 'version': 1},
                          'requestNr': 1,
                          'errorcode': 0,
                          'content': {'type': 1, 'attachment': {}},
                          'attachResponse': True}]

        part2 = {}
        part2['Trustor'] = {'Provider': 'cpp', 'ID': self._client_id, 'Type': 'FI-AIR-AND'}
        part2['Trustee'] = {'Provider': 'cpp', 'ID': self._device_id, 'Type': 'AirPurifier'}
        part2['PairingData'] = {'Secret': pair_secret, 'MatchIPAddress': False, 'RequestTtl': 5}
        part2['Relationship'] = {'Type': 'di-comm', 'Permissions': ['Response', 'Change'], 'AllowDelegation': False, 'Metadata': '', 'Ttl': 1000000000}

        req_body = self._multi_part(part1, part2)
        if self._debug:
            print(req_body)

        req = urllib.request.Request(url=url, data=req_body, method='POST')
        req.add_header('Authorization', auth)
        req.add_header('Content-Type', 'multipart/mixed; boundary="ICPMimeBoundary"')
        with urllib.request.urlopen(req) as response:
            content_type = response.getheader('Content-Type')
            body = response.read().decode('ascii')
            parts = "Content-Type: %s\n\n%s" % (content_type, body)

        if self._debug:
            print(parts)
        msg = email.message_from_string(parts)
        if not msg.is_multipart():
            raise Exception('Expected multipart response')
        for part in msg.get_payload():
            if part.get_content_type() == 'application/cb-message':
                continue
            p = json.loads(part.get_payload(decode=False))
            rel_status = p['RelationshipStatus']
            print('Relationship status: {}'.format(rel_status))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('device_id', help='Purifier device ID')
    parser.add_argument('-d', '--debug', help='show debug output', action='store_true')
    parser.add_argument('--om', help='set fan speed', choices=['1','2','3','s','t'])
    parser.add_argument('--pwr', help='power on/off', choices=['0','1'])
    parser.add_argument('--mode', help='set mode', choices=['P','A','S','M','B','N'])
    parser.add_argument('--rhset', help='set target humidity', choices=['40','50','60','70'])
    parser.add_argument('--func', help='set function', choices=['P','PH'])
    parser.add_argument('--aqil', help='set light brightness', choices=['0','25','50','75','100'])
    parser.add_argument('--uil', help='set button lights on/off', choices=['0','1'])
    parser.add_argument('--ddp', help='set indicator pm2.5/IAI', choices=['0','1'])
    parser.add_argument('--dt', help='set timer', choices=['0','1','2','3','4','5'])
    parser.add_argument('--pair', help='pair with the device having the given IP')
    args = parser.parse_args()

    c = CloudClient(args.device_id, args.debug)
    c.load_credentials()
    if args.pair:
        c.pair(args.pair)

    values = {}
    if args.om:
        values['om'] = args.om
    if args.pwr:
        values['pwr'] = args.pwr
    if args.mode:
        values['mode'] = args.mode
    if args.rhset:
        values['rhset'] = int(args.rhset)
    if args.func:
        values['func'] = args.func
    if args.aqil:
        values['aqil'] = int(args.aqil)
    if args.ddp:
        values['ddp'] = args.ddp
    if args.uil:
        values['uil'] = args.uil
    if args.dt:
        values['dt'] = int(args.dt)

    if values:
        c.set_values(values)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import argparse
import sys
import pprint

from pyairctrl.status_transformer import STATUS_TRANSFORMER
from pyairctrl.coap_client import CoAPAirClient
from pyairctrl.http_client import HTTPAirClient
from pyairctrl.plain_coap_client import PlainCoAPAirClient


class CoAPCli:
    def __init__(self, host, port=5683, debug=False):
        self._client = CoAPAirClient(host, port, debug)

    def _get_info_for_key(self, key, current_value):
        if key in STATUS_TRANSFORMER:
            info = STATUS_TRANSFORMER[key]
            if not info[1] is None:
                current_value = info[1](current_value)
                if current_value is None:
                    return None
            return info[0].format(current_value)

        return "{}: {}".format(key, current_value)

    def _dump_status(self, status, debug=False):
        if debug:
            print("Raw status:")
            pprint.pprint(status)
        for key in status:
            current_value = status[key]
            name_and_value = self._get_info_for_key(key, current_value)
            if name_and_value is None:
                continue

            print(
                "[{key}]\t{name_and_value}".format(
                    key=key, name_and_value=name_and_value
                ).expandtabs(30)
            )

    def get_status(self, debug=False):
        status = self._client.get_status(debug)
        if status is not None:
            return self._dump_status(status, debug=debug)

    def set_values(self, values, debug=False):
        self._client.set_values(values, debug)

    def get_firmware(self):
        status = self._client.get_firmware()
        if status is None:
            print("No version-info found")
            return

        print(
            self._get_info_for_key(
                "swversion", status["swversion"] if "swversion" in status else "nA"
            )
        )
        print(self._get_info_for_key("ota", status["ota"] if "ota" in status else "nA"))

    def get_filters(self):
        status = self._client.get_filters()
        if status is None:
            print("No version-info found")
            return

        if "fltsts0" in status:
            print(self._get_info_for_key("fltsts0", status["fltsts0"]))
        if "fltsts1" in status:
            print(self._get_info_for_key("fltsts1", status["fltsts1"]))
        if "fltsts2" in status:
            print(self._get_info_for_key("fltsts2", status["fltsts2"]))
        if "wicksts" in status:
            print(self._get_info_for_key("wicksts", status["wicksts"]))

    def get_wifi(self):
        print(
            "Getting wifi credentials is currently not supported when using CoAP. Use the app instead."
        )

    def set_wifi(self, ssid, pwd):
        print(
            "Setting wifi credentials is currently not supported when using CoAP. Use the app instead."
        )


class HTTPAirCli:
    @staticmethod
    def ssdp(timeout=1, repeats=3, debug=False):
        response = HTTPAirClient.ssdp(timeout, repeats)
        if debug:
            pprint.pprint(response)
        return response

    def __init__(self, host, debug=True):
        self._client = HTTPAirClient(host, debug)

    def set_values(self, values, debug=False):
        try:
            values = self._client.set_values(values)
            self._dump_status(values, debug=debug)
        except urllib.error.HTTPError as e:
            print("Error setting values (response code: {})".format(e.code))

    def set_wifi(self, ssid, pwd):
        values = {}
        if ssid:
            values["ssid"] = ssid
        if pwd:
            values["password"] = pwd
        pprint.pprint(values)

        wifi = self._client.set_wifi(ssid, pwd)
        pprint.pprint(wifi)

    def _dump_status(self, status, debug=False):
        if debug:
            pprint.pprint(status)
            print()

        if "pwr" in status:
            pwr = status["pwr"]
            pwr_str = {"1": "ON", "0": "OFF"}
            pwr = pwr_str.get(pwr, pwr)
            print("[pwr]   Power: {}".format(pwr))
        if "pm25" in status:
            pm25 = status["pm25"]
            print("[pm25]  PM25: {}".format(pm25))
        if "rh" in status:
            rh = status["rh"]
            print("[rh]    Humidity: {}".format(rh))
        if "rhset" in status:
            rhset = status["rhset"]
            print("[rhset] Target humidity: {}".format(rhset))
        if "iaql" in status:
            iaql = status["iaql"]
            print("[iaql]  Allergen index: {}".format(iaql))
        if "temp" in status:
            temp = status["temp"]
            print("[temp]  Temperature: {}".format(temp))
        if "func" in status:
            func = status["func"]
            func_str = {"P": "Purification", "PH": "Purification & Humidification"}
            func = func_str.get(func, func)
            print("[func]  Function: {}".format(func))
        if "mode" in status:
            mode = status["mode"]
            mode_str = {
                "P": "auto",
                "A": "allergen",
                "S": "sleep",
                "M": "manual",
                "B": "bacteria",
                "N": "night",
                "T": "turbo",
            }
            mode = mode_str.get(mode, mode)
            print("[mode]  Mode: {}".format(mode))
        if "om" in status:
            om = status["om"]
            om_str = {"s": "silent", "t": "turbo"}
            om = om_str.get(om, om)
            print("[om]    Fan speed: {}".format(om))
        if "aqil" in status:
            aqil = status["aqil"]
            print("[aqil]  Light brightness: {}".format(aqil))
        if "uil" in status:
            uil = status["uil"]
            uil_str = {"1": "ON", "0": "OFF"}
            uil = uil_str.get(uil, uil)
            print("[uil]   Buttons light: {}".format(uil))
        if "ddp" in status:
            ddp = status["ddp"]
            ddp_str = {"1": "PM2.5", "0": "IAI", "2": "Gas"}
            ddp = ddp_str.get(ddp, ddp)
            print("[ddp]   Used index: {}".format(ddp))
        if "wl" in status:
            wl = status["wl"]
            print("[wl]    Water level: {}".format(wl))
        if "cl" in status:
            cl = status["cl"]
            print("[cl]    Child lock: {}".format(cl))
        if "dt" in status:
            dt = status["dt"]
            if dt != 0:
                print("[dt]    Timer: {} hours".format(dt))
        if "dtrs" in status:
            dtrs = status["dtrs"]
            if dtrs != 0:
                print("[dtrs]  Timer: {} minutes left".format(dtrs))
        if "err" in status:
            err = status["err"]
            if err != 0:
                err_str = {
                    49408: "no water",
                    32768: "water tank open",
                    49155: "pre-filter must be cleaned",
                }
                err = err_str.get(err, err)
                print("-" * 20)
                print("Error: {}".format(err))

    def get_status(self, debug=False):
        status = self._client.get_status()
        self._dump_status(status, debug=debug)

    def get_wifi(self):
        wifi = self._client.get_wifi()
        pprint.pprint(wifi)

    def get_firmware(self):
        firmware = self._client.get_firmware()
        pprint.pprint(firmware)

    def get_filters(self):
        filters = self._client.get_filters()
        print("Pre-filter and Wick: clean in {} hours".format(filters["fltsts0"]))
        if "wicksts" in filters:
            print("Wick filter: replace in {} hours".format(filters["wicksts"]))
        print("Active carbon filter: replace in {} hours".format(filters["fltsts2"]))
        print("HEPA filter: replace in {} hours".format(filters["fltsts1"]))


class PlainCoAPAirCli:
    def __init__(self, host, port=5683):
        self._client = PlainCoAPAirClient(host, port)

    def _dump_status(self, status, debug=False):
        if debug:
            print("Raw status: " + str(status))
        if "name" in status:
            name = status["name"]
            print("[name]        Name: {}".format(name))
        if "modelid" in status:
            modelid = status["modelid"]
            print("[modelid]     ModelId: {}".format(modelid))
        if "swversion" in status:
            swversion = status["swversion"]
            print("[swversion]   Version: {}".format(swversion))
        if "StatusType" in status:
            statustype = status["StatusType"]
            print("[StatusType]  StatusType: {}".format(statustype))
        if "ota" in status:
            ota = status["ota"]
            print("[ota]         Over the air updates: {}".format(ota))
        if "Runtime" in status:
            runtime = status["Runtime"]
            print(
                "[Runtime]     Runtime: {} hours".format(
                    round(((runtime / (1000 * 60 * 60)) % 24), 2)
                )
            )
        if "pwr" in status:
            pwr = status["pwr"]
            pwr_str = {"1": "ON", "0": "OFF"}
            pwr = pwr_str.get(pwr, pwr)
            print("[pwr]         Power: {}".format(pwr))
        if "pm25" in status:
            pm25 = status["pm25"]
            print("[pm25]        PM25: {}".format(pm25))
        if "rh" in status:
            rh = status["rh"]
            print("[rh]          Humidity: {}".format(rh))
        if "rhset" in status:
            rhset = status["rhset"]
            print("[rhset]       Target humidity: {}".format(rhset))
        if "iaql" in status:
            iaql = status["iaql"]
            print("[iaql]        Allergen index: {}".format(iaql))
        if "temp" in status:
            temp = status["temp"]
            print("[temp]        Temperature: {}".format(temp))
        if "func" in status:
            func = status["func"]
            func_str = {"P": "Purification", "PH": "Purification & Humidification"}
            func = func_str.get(func, func)
            print("[func]        Function: {}".format(func))
        if "mode" in status:
            mode = status["mode"]
            mode_str = {
                "P": "auto",
                "A": "allergen",
                "S": "sleep",
                "M": "manual",
                "B": "bacteria",
                "N": "night",
                "T": "turbo",
            }
            mode = mode_str.get(mode, mode)
            print("[mode]        Mode: {}".format(mode))
        if "om" in status:
            om = status["om"]
            om_str = {"s": "silent", "t": "turbo"}
            om = om_str.get(om, om)
            print("[om]          Fan speed: {}".format(om))
        if "aqil" in status:
            aqil = status["aqil"]
            print("[aqil]        Light brightness: {}".format(aqil))
        if "uil" in status:
            uil = status["uil"]
            uil_str = {"1": "ON", "0": "OFF"}
            uil = uil_str.get(uil, uil)
            print("[uil]         Buttons light: {}".format(uil))
        if "ddp" in status:
            ddp = status["ddp"]
            ddp_str = {"3": "Humidity", "1": "PM2.5", "2": "Gas", "0": "IAI"}
            ddp = ddp_str.get(ddp, ddp)
            print("[ddp]         Used index: {}".format(ddp))
        if "wl" in status:
            wl = status["wl"]
            print("[wl]          Water level: {}".format(wl))
        if "cl" in status:
            cl = status["cl"]
            print("[cl]          Child lock: {}".format(cl))
        if "dt" in status:
            dt = status["dt"]
            if dt != 0:
                print("[dt]          Timer: {} hours".format(dt))
        if "dtrs" in status:
            dtrs = status["dtrs"]
            if dtrs != 0:
                print("[dtrs]        Timer: {} minutes left".format(dtrs))
        if "fltsts0" in status:
            fltsts0 = status["fltsts0"]
            print(
                "[fltsts0]     Pre-filter and Wick: clean in {} hours".format(fltsts0)
            )
        if "fltsts1" in status:
            fltsts1 = status["fltsts1"]
            print("[fltsts1]     HEPA filter: replace in {} hours".format(fltsts1))
        if "fltsts2" in status:
            fltsts2 = status["fltsts2"]
            print(
                "[fltsts2]     Active carbon filter: replace in {} hours".format(
                    fltsts2
                )
            )
        if "wicksts" in status:
            wicksts = status["wicksts"]
            print("[wicksts]     Wick filter: replace in {} hours".format(wicksts))
        if "err" in status:
            err = status["err"]
            if err != 0:
                err_str = {
                    49408: "no water",
                    32768: "water tank open",
                    49155: "pre-filter must be cleaned",
                }
                err = err_str.get(err, err)
                print("-" * 20)
                print("[ERROR] Message: {}".format(err))

    def set_values(self, values, debug=False):
        self._client.set_values(values, debug)

    def get_status(self, debug=False):
        status = self._client.get_status(debug)
        return self._dump_status(status, debug=debug)

    def get_wifi(self):
        print(
            "Getting wifi credentials is currently not supported when using CoAP. Use the app instead."
        )

    def set_wifi(self, ssid, pwd):
        print(
            "Setting wifi credentials is currently not supported when using CoAP. Use the app instead."
        )

    def get_firmware(self):
        status = self._client.get_firmware()
        print("Software version: {}".format(status["swversion"]))
        print("Over the air updates: {}".format(status["ota"]))

    def get_filters(self):
        status = self._client.get_filters()
        print("Pre-filter and Wick: clean in {} hours".format(status["fltsts0"]))
        print("HEPA filter: replace in {} hours".format(status["fltsts1"]))
        print("Active carbon filter: replace in {} hours".format(status["fltsts2"]))
        print("Wick filter: replace in {} hours".format(status["wicksts"]))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ipaddr", help="IP address of air purifier")
    parser.add_argument(
        "--protocol",
        help="set the communication protocol",
        choices=["http", "coap", "plain_coap"],
        default="http",
    )
    parser.add_argument("-d", "--debug", help="show debug output", action="store_true")
    parser.add_argument("--om", help="set fan speed", choices=["1", "2", "3", "s", "t", "a"])
    parser.add_argument("--pwr", help="power on/off", choices=["0", "1"])
    parser.add_argument(
        "--mode", help="set mode", choices=["P", "A", "S", "M", "B", "N", "T"]
    )
    parser.add_argument(
        "--rhset", help="set target humidity", choices=["40", "50", "60", "70"]
    )
    parser.add_argument("--func", help="set function", choices=["P", "PH"])
    parser.add_argument(
        "--aqil", help="set light brightness", choices=["0", "25", "50", "75", "100"]
    )
    parser.add_argument("--uil", help="set button lights on/off", choices=["0", "1"])
    parser.add_argument(
        "--ddp", help="set indicator IAI/PM2.5/Gas/Humidity", choices=["0", "1", "2", "3"]
    )
    parser.add_argument(
        "--dt",
        help="set timer",
        choices=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"],
    )
    parser.add_argument("--cl", help="set child lock", choices=["True", "False"])
    parser.add_argument("--wifi", help="read wifi options", action="store_true")
    parser.add_argument("--wifi-ssid", help="set wifi ssid")
    parser.add_argument("--wifi-pwd", help="set wifi password")
    parser.add_argument("--firmware", help="read firmware", action="store_true")
    parser.add_argument("--filters", help="read filters status", action="store_true")
    args = parser.parse_args()

    if args.ipaddr:
        devices = [{"ip": args.ipaddr}]
    else:
        if args.protocol in ["coap", "plain_coap"]:
            print(
                "Autodetection is not supported when using CoAP. Use --ipaddr to set an IP address."
            )
            sys.exit(1)

        devices = HTTPAirCli.ssdp(debug=args.debug)
        if not devices:
            print(
                "Air purifier not autodetected. Try --ipaddr option to force specific IP address."
            )
            sys.exit(1)

    for device in devices:
        if args.protocol == "http":
            c = HTTPAirCli(device["ip"])
        elif args.protocol == "plain_coap":
            c = PlainCoAPAirCli(device["ip"])
        elif args.protocol == "coap":
            c = CoAPCli(device["ip"], debug=args.debug)

        if args.wifi:
            c.get_wifi()
            sys.exit(0)
        if args.firmware:
            c.get_firmware()
            sys.exit(0)
        if args.wifi_ssid or args.wifi_pwd:
            c.set_wifi(args.wifi_ssid, args.wifi_pwd)
            sys.exit(0)
        if args.filters:
            c.get_filters()
            sys.exit(0)

        values = {}
        if args.om:
            values["om"] = args.om
        if args.pwr:
            values["pwr"] = args.pwr
        if args.mode:
            values["mode"] = args.mode
        if args.rhset:
            values["rhset"] = int(args.rhset)
        if args.func:
            values["func"] = args.func
        if args.aqil:
            values["aqil"] = int(args.aqil)
        if args.ddp:
            values["ddp"] = args.ddp
        if args.uil:
            values["uil"] = args.uil
        if args.dt:
            values["dt"] = int(args.dt)
        if args.cl:
            values["cl"] = args.cl == "True"

        if values:
            c.set_values(values, debug=args.debug)
        else:
            c.get_status(debug=args.debug)


if __name__ == "__main__":
    main()



"""Plain CoAP Air Client."""

# pylint: disable=invalid-name, missing-class-docstring, missing-function-docstring

import json
import logging
import os
import socket
import struct
import sys
import time

from coapthon import defines
from coapthon.client.helperclient import HelperClient
from coapthon.messages.request import Request
from coapthon.utils import generate_random_token


class NotSupportedException(Exception):
    pass


class PlainCoAPAirClient:
    def __init__(self, host, port=5683):
        self.coapthon_logger = logging.getLogger("coapthon")
        self.coapthon_logger.setLevel("WARN")
        self.server = host
        self.port = port

    def _create_coap_client(self, host, port):
        return HelperClient(server=(host, port))

    def _send_over_socket(self, destination, packet):
        protocol = socket.getprotobyname("icmp")
        if os.geteuid() == 0:
            s = socket.socket(socket.AF_INET, socket.SOCK_RAW, protocol)
        else:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, protocol)
        try:
            s.sendto(packet, (destination, 0))
        except OSError:  # That fixes a mac os bug for me: OSError: [Errno 22] Invalid argument
            pass
        finally:
            s.close()

    def _get(self):
        path = "/sys/dev/status"
        try:
            client = self._create_coap_client(self.server, self.port)
            self._send_hello_sequence(client)
            request = client.mk_request(defines.Codes.GET, path)
            request.destination = server = (self.server, self.port)
            request.type = defines.Types["ACK"]
            request.token = generate_random_token(4)
            request.observe = 0
            response = client.send_request(request, None, 2)
        finally:
            client.stop()

        if response:
            return json.loads(response.payload)["state"]["reported"]
        else:
            return {}

    def _set(self, key, value):
        path = "/sys/dev/control"
        try:
            client = self._create_coap_client(self.server, self.port)
            self._send_hello_sequence(client)
            payload = {"state": {"desired": {key: value}}}
            response = client.post(path, json.dumps(payload))
            return response.payload == '{"status":"success"}'
        finally:
            client.stop()

    def _send_hello_sequence(self, client):
        ownIp = self._get_ip()

        header = self._create_icmp_header()
        data = self._create_icmp_data(ownIp, self.port, self.server, self.port)
        packet = header + data
        packet = self._create_icmp_header(self._checksum_icmp(packet)) + data

        self._send_over_socket(self.server, packet)

        # that is needed to give device time to open coap port, otherwise it may not respond properly
        time.sleep(0.5)

        request = Request()
        request.destination = server = (self.server, self.port)
        request.code = defines.Codes.EMPTY.number
        client.send_empty(request)

    def _get_ip(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(("10.255.255.255", 1))
            ip = s.getsockname()[0]
        except:
            ip = "127.0.0.1"
        finally:
            s.close()
        return ip

    def _checksum_icmp(self, source_string):
        countTo = (int(len(source_string) / 2)) * 2
        sum = 0
        count = 0
        loByte = 0
        hiByte = 0
        while count < countTo:
            if sys.byteorder == "little":
                loByte = source_string[count]
                hiByte = source_string[count + 1]
            else:
                loByte = source_string[count + 1]
                hiByte = source_string[count]
            sum = sum + (hiByte * 256 + loByte)
            count += 2

        if countTo < len(source_string):  # Check for odd length
            loByte = source_string[len(source_string) - 1]
            sum += loByte

        sum &= 0xFFFFFFFF
        sum = (sum >> 16) + (sum & 0xFFFF)
        sum += sum >> 16
        answer = ~sum & 0xFFFF
        answer = socket.htons(answer)
        return answer

    def _create_icmp_header(self, checksum=0):
        ICMP_TYPE = 3
        ICMP_CODE = 3
        UNUSED = 0
        CHECKSUM = checksum
        header = struct.pack("!BBHI", ICMP_TYPE, ICMP_CODE, CHECKSUM, UNUSED)
        return header

    def _checksum_tcp(self, pkt):
        return 0  # looks like its irrelevant what we send here

    def _create_tcp_data(self, srcIp, dstIp, checksum=0):
        ip_version = 4
        ip_vhl = 5

        ip_ver = (ip_version << 4) + ip_vhl

        # Differentiate Service Field
        ip_dsc = 0
        ip_ecn = 0

        ip_dfc = (ip_dsc << 2) + ip_ecn

        # Total Length
        ip_tol = 214

        # Identification
        ip_idf = 6190

        # Flags
        ip_rsv = 0
        ip_dtf = 0
        ip_mrf = 0
        ip_frag_offset = 0

        ip_flg = (ip_rsv << 7) + (ip_dtf << 6) + (ip_mrf << 5) + (ip_frag_offset)

        # Time to live
        ip_ttl = 255

        # Protocol
        ip_proto = socket.IPPROTO_UDP

        # Check Sum
        ip_chk = checksum

        # Source Address
        ip_saddr = socket.inet_aton(srcIp)

        # Destination Address
        ip_daddr = socket.inet_aton(dstIp)
        tcp = struct.pack(
            "!BBHHHBBH4s4s",
            ip_ver,  # IP Version
            ip_dfc,  # Differentiate Service Feild
            ip_tol,  # Total Length
            ip_idf,  # Identification
            ip_flg,  # Flags
            ip_ttl,  # Time to leave
            ip_proto,  # protocol
            ip_chk,  # Checksum
            ip_saddr,  # Source IP
            ip_daddr,  # Destination IP
        )
        return tcp

    def _create_udp_data(self, srcPort, dstPort):
        sport = srcPort
        dport = dstPort
        length = 194
        checksum = 0
        udp = struct.pack("!HHHH", sport, dport, length, checksum)
        return udp

    def _create_icmp_data(self, srcIp, srcPort, dstIp, dstPort):
        return self._create_tcp_data(srcIp, dstIp) + self._create_udp_data(
            srcPort, dstPort
        )

    def set_values(self, values, debug=False):
        if debug:
            self.coapthon_logger.setLevel("DEBUG")
        result = True
        for key in values:
            result = result and self._set(key, values[key])

        return result

    def get_status(self, debug=False):
        if debug:
            self.coapthon_logger.setLevel("DEBUG")
        status = self._get()
        return status

    def get_wifi(self):
        raise NotSupportedException

    def set_wifi(self, ssid, pwd):
        raise NotSupportedException

    def get_firmware(self):
        status = self._get()
        # TODO Really transmit full status here?
        return status

    def get_filters(self):
        status = self._get()
        # TODO Really transmit full status here?
        return status
